
require "Class"
require "List"

HogeAI = Class {
	initialize = function (self, myid)
		self.id = myid
		self.actor = Homun.new(self.id)
		self.owner = self.actor:owner()
		self.resCmds = List.new() -- 予約コマンド (関数のリスト)
		self.state = "HUKAWA_ADDICT"

		self.steps = List.new()
		self.friends = List.new()

		self.ignoreMonsters = loadstring("return "..Config.AI.ignore)()
		info("IgnoreMonsters: "..self.ignoreMonsters:join())
	end,

	-- called by RO per 140ms
	act = function (self)
		-- process timer
		local time = os.time()
		if not self.prevTime or
		   self.prevTime ~= time then

			if math.mod(time, 5) == 0 then 
				self:checkStack()
			end
			self.prevTime = time
		end

		-- process messages
		local msg = GetMsg(self.id)
		local rmsg = GetResMsg(self.id)
		if msg[1] ~= NONE_CMD then
			self.resCmds:clear()
			self:process(msg)
		else
			if rmsg[1] ~= NONE_CMD then
				self.resCmds:push(function () 
					return self:process(rmsg) 
				end)
			end
		end

		-- process reserved commands
		-- true を返したら後続処理をやめる
		if self.resCmds:size() > 0 and self.resCmds:shift()() then
			return
		end

		-- process update state
		local state = self.state
		local name = self.state .. "_STATE"
		if self[name] then self[name](self) end
		if state ~= self.state then
			d(state.." -> "..self.state)
		end

	end,

	process = function (self, msg)
		-- dispatch
		local name = "on" .. COMMAND[msg[1]]
		d("process command: " .. name, unpack(msg))
		if self[name] then return self[name](self, unpack(msg)) end
	end,

	onMOVE_CMD = function (self, id, x, y)
		if self.owner:isSit() then
			Actor.playersByPosition(x, y):each(function (player)
				if self.friends:isInclude(player) then
					d(string.format("Remove %s (%s) as friends list.", player.id, player:jobName()))
					self.friends:delete(player)
				else
					d(string.format("Add %s (%s) as friends list.", player.id, player:jobName()))
					self.friends:push(player)
				end
			end)
		end
		self:stepCommand(x, y)
		self.actor:move(x, y)
	end,

	onFOLLOW_CMD = function (self)
		-- Alt-T での切り替え
		if self.state ~= "FORCE_FOLLOW" then
			self.state = "FORCE_FOLLOW"
			d("-> "..self.state)
		else
			self.state = "IDLE"
			d("-> "..self.state)
		end
	end,

	onATTACK_OBJECT_CMD = function (self, id, target)
		self:attack(Actor.new(target))
		d("-> "..self.state)
	end,

	onSKILL_OBJECT_CMD = function (self, id, level, skill, target)
		target = Actor.new(target)
		self.actor:skillObject(level, skill, target)
	end,


	-- 完全なアイドル状態。敵を無視してフォロー
	FORCE_FOLLOW_STATE = function (self)
		-- 主人が攻撃しているときは援護
		if self:attackOwnerTarget() then return true end

		if self.actor:distanceFrom(self.owner) > 2 then
			-- self.actor:moveToNear(self.owner)
			self.actor:move(self.owner:position())
		end
	end,

	-- アイドル状態。ただしアクティブ
	IDLE_STATE = function (self)
		-- search mob
		if self:searchAndAttack() then return end

		-- follow
		local distance = self.actor:distanceFrom(self.owner)
		if (distance > 3 or distance == -1) then
			self.state = "FOLLOW"
			return
		end
	end,

	CHASE_STATE = function (self)
		try(function ()
			-- 主人からの距離で見る
			if self.owner:distanceFrom(self.target) > 8 then
				d(string.format("Target is too far: %s %s", self.target.id, self.target.info.NAME))
				self.state = self.prevState
				self.target = nil
				return
			end

			-- 動いたらターゲット位置を変更
			local px, py = self.target.px, self.target.py
			local x, y = self.target:position()
			if (px ~= x or py ~= y) then
				self.target.px, self.target.py = x, y -- 位置を記憶
				self.actor:moveToNear(self.target)
				return
			end

			if self.actor:isInAttackSight(self.target) then
				self.state = "ATTACK"
				return
			end
		end).catch(function (e)
			d(e)
			self.state = self.prevState
			self.target = nil
		end)
	end,

	ATTACK_STATE = function (self)
		if not self.target then return end

		try(function ()
			if not self.actor:isInSight(self.target) then
				d(string.format("Target is out of sight: %s %s", self.target.id, self.target.info.NAME))
				self.state = self.prevState
				self.target = nil
				return
			end

			if not self.actor:isInAttackSight(self.target) then
				self.state = "CHASE"
				self.actor:moveToNear(self.target)
				return
			end
			
			if (
			    (self.actor.atk and self.actor.atk * 5 < self.target:maxHp()) and
			     not self.target.isAttacked and
			     not self.target:isPlant()
			   ) and (
			    (self.actor:spRate() > 80) or
			    (self.actor:spRate() > 50 and Actor.monsters():size() > 2) or
			    false
			   )
			   then

				d"Try to use skill"
				self.actor:skillObject(5, SKILL_MOONLIGHT, self.target)
			end

			self.actor:attack(self.target)
			local x, y = self.target:position()
			self.actor:move(x + math.random(3) - 2, y + math.random(3) - 2)
			-- self.actor:attack(self.target)

			self.target.isAttacked = true

		end).catch(function (e)
			d(e)
			self.state = self.prevState
			self.target = nil
		end)

		if self.actor:distanceFrom(self.owner) > 8 then
			self.actor:moveToOwner()
			return
		end
	end,

	-- 主人を追従
	FOLLOW_STATE = function (self)
		if self:searchAndAttack() then return end

		if self.actor:distanceFrom(self.owner) <= 3 then
			self.state = "IDLE"
		elseif self.actor:isStand() then
			self.actor:moveToNear(self.owner)
		end
	end,

	HUKAWA_ADDICT_STATE = function (self)
		local Hukawa = Actor.new(51105)
		if not Hukawa:exists() then
			d"Hukawa does not exists"
			self.state = "FORCE_FOLLOW"
			return
		end
		self:stepMotion(44, 180)
	end,

	stepCommand = function (self, x, y)
		local ox, oy = self.owner:position()
		d(string.format("step: owner: %d, %d - target: %d, %d", ox, oy, x, y))

		-- クリック位置の判定
		local command = (function ()
			if oy == y and ox == x + 1 then return "L" end
			if oy == y and ox == x - 1 then return "R" end
			if ox == x and oy == y + 1 then return "D" end
			if ox == x and oy == y - 1 then return "U" end
			return nil
		end)(x, y)

		if command then
			d("step: "..command)
			self.steps:push(command)
			self.steps = self.steps:last(3)

			s = self.steps
			d(s:join(""))
			if s:join("") == "LRD" then
				self.steps:clear()
				d"Reload AI"
				reload()
				return true
			end
		else
			self.steps:clear()
		end
	end,

	-- 主人のタゲを処理
	attackOwnerTarget = function (self)
		local target = self.owner:target()
		if target:exists() then
			self:attack(target)
			return true
		end
	end,
	
	searchAndAttack = function (self)
		if self:attackOwnerTarget() then return true end

		local pick = Actor.monsters():sortBy(function (i)
			local rate = self.actor:distanceFrom(i)
			if rate > 8 then rate = 99999 end

			local target = i:target()
			if target == self.actor or
			   target == self.owner then
				rate = rate - 1000
			else
				if self.ignoreMonsters:isInclude(i:homunType()) then rate = 99999 end
				if i:isBioPlant() then rate = 99999 end
				if i:isPlant() then rate = 99999 end

				-- 横殴り防止
				-- タゲが決まっていたらやめる
				if target:exists() then rate = 99999 end
--[[
				-- 10 セル以内に誰かいたらやめる
				if not Actor.players():select(function (a)
					if a == self.actor then return false end
					if a == self.owner then return false end
					if self.friends:isInclude(a) then return false end
					-- TODO: 職業による射程を考慮
					if a:distanceFrom(i) < 15 then return true end
				end):isEmpty() then
					rate = rate + 10000
				end
]]
			end
			i.rate = rate
			return rate
		end)

		if not pick:isEmpty() and
		   pick[1].rate < 10000 then
			d(string.format(
				"Set target -> %s: distance: %s, rate: %s",
				pick[1].id,
				self.owner:distanceFrom(pick[1]),
				pick[1].rate
			))
			self:attack(pick[1])
		end
	end,

	checkStack = function (self)
		for i, v in ipairs({"HUKAWA_ADDICT", "SLEEP"}) do
			if self.state == v then return end
		end
		if self.actor:distanceFrom(self.owner) > 3 then
			d"checkStack"

			local x, y = self.actor:position()
			if self.prevPos then
				local px, py = self.prevPos.x, self.prevPos.y

				-- 前回と同じ場所にいたら
				if px == x and
				   py == y then
					if not self.state == "FORCE_FOLLOW" then
						d"Stacked? state changing -> return to owner position"
					end
					-- self.actor:move(self.owner:position())
					self.actor:moveToOwner()
				end
			end

			self.prevPos = { x = x, y = y }
		end
	end,

	attack = function (self, target)
		-- ATTACK 後に prevState に復帰する
		self.prevState = self.state
		self.target = target
		self.state = "CHASE"
		target:getMonsterInfo()
		info(string.format(
			"Will attack->%s %s (%s), HP:%s",
			target.id,
			target.info.NAME,
			target:homunType(),
			target:maxHp()))
	end,

	-- いわゆるくるくる ox, oy を中心にして一周する
	-- 省略時はオーナー周辺
	stepMotion = function (self, ox, oy, steps)
		steps = steps or {
			{ x =  0, y = 1  },
			{ x =  1, y = 1  },
			{ x =  1, y = -1 },
			{ x = -1, y = -1 },
			{ x = -1, y = 1  },
			{ x =  0, y = 1  },
		}
		if not ox then ox, oy = self.owner:position() end

		d(string.format("stepMotion -> [%s, %s]", ox, oy))

		local co; co = coroutine.create(function (init)
			for i, v in pairs(steps) do
				local tx, ty = ox + v.x, oy + v.y -- target
				while true do
					self.actor:move(tx, ty)

					local x, y = self.actor:position()
					-- 移動不可なセル付近で実行すると一生かえってこないだろう
					if x == tx and y == ty then
						break
					else
						-- コマンドを予約して
						self.resCmds:push(function () return coroutine.resume(co) end)
						-- 処理を一旦戻す 
						-- (次の AI 呼び出しのコマンドプロセスで、ここから再開する)
						coroutine.yield(true)
					end
				end
			end
		end)
		coroutine.resume(co)
	end,
}
