--[[
Homun class
===========

DESCRIPTION
-----------

class for my own homunculus

自分自身のホムンクルスに対するクラスです。
Actor クラスを継承しており、Actor のメソッドが呼べます。

NOTE
----

このクラスは singleton として使うことを想定しています。
(ただし singleton としての実装はしていません)

]]

require "Class"
require "Actor"

Homun = Class { super = Actor,
	initialize = function (self, id)
		self.super.initialize(self, id)

		if Config and Config.Homun then
			for k, v in pairs(Config.Homun) do
				self[k] = loadstring("return "..v)()
			end
		end
	end,

	owner = function (self)
		return Actor.new(GetV(V_OWNER, self.id))
	end,

	move = function (self, x, y)
		Move(self.id, x, y)
	end,

	moveToOwner = function (self)
		MoveToOwner(self.id)
	end,

	-- target の周囲 8 セルで一番近いセルに移動する
	moveToNear = function (self, target)
		local sx, sy = self:position()
		local tx, ty = target:position()
		local cand = List.new({x =  1, y =  1},
		                      {x =  1, y =  0},
		                      {x =  0, y =  1},
		                      {x = -1, y = -1},
		                      {x = -1, y =  0},
		                      {x =  0, y = -1},
		                      {x =  0, y  = 0})
		local res = cand:map(function (p)
			return {x = tx + p.x, y = ty + p.y}
		end):sortBy(function (p)
			return math.floor(math.sqrt((sx - p.x) ^ 2 + (sy - p.y) ^ 2))
		end)[1]
		self:move(res.x, res.y)
	end,

	attack = function (self, target)
		Attack(self.id, target.id)
	end,

	-- use skill
	skillObject = function (self, level, skill, target)
		SkillObject(self.id, level, skill, target.id)
	end,

	skillGround = function (self, level, skill, x, y)
		SkillGround(self.id, level, skill, x, y)
	end,
}

