--[[
INI class 
===========

DESCRIPTION
-----------

ini ファイルを読み書きするためのクラス

SINOPSYS
--------


	INI.new({
		SectionName = {
			key = "value"
		}
	}):write("hoge.ini")

	ini = INI.new():read("hoge.ini")
	ini.SectionName.key

NOTE
----

書きこみの際は順番が保持されません。

]]


require "Class"

INI = Class {
	initialize = function (self, table)
		if table then
			for i, v in pairs(table) do
				i = tostring(i)
				self[i] = {}
				for j, w in pairs(v) do
					self[i][j] = tostring(w)
				end
			end
		end
	end,

	read = function (self, filename)
		local sect  = nil
		local name  = nil
		for line in io.lines(filename) do
			line = string.gsub(line, "%s*;.*", "")
			s, e, n  = string.find(line, "%[(.+)%]")
			if s then
				if sect then self[name] = sect end
				name = n
				sect = {}
			else
				s, e, k, v = string.find(line, "(%w+)=(.*)")
				if s then
					sect[k] = v
				end
			end
		end
		if sect then self[name] = sect end
		return self
	end,

	write = function (self, filename)
		local f = io.open(filename, "w")
		for i, v in pairs(self) do
			f:write(string.format("[%s]\n", i))
			for j, w in pairs(v) do
				f:write(string.format("%s=%s\n", j, w))
			end
			f:write("\n")
		end
		f:close()
	end,
}


