#!/opt/local/bin/lua


require "Class"
require "Test"

Point = Class {
	__tostring = function (self)
		return "#<Class Point>"
	end,

	initialize = function (self, x, y)
		self.x = x
		self.y = y
	end,

	point = function (self)
		return self.x, self.y
	end,

	__add = function (a, b)
		return Point.new(a.x + b.x, a.y + b.y)
	end,

	__eq = function (a, b)
		return (a.x == b.x)
	end,
}

Point3D = Class { super = Point,
	__tostring = function (self)
		return "#<Class Point3D>"
	end,

	initialize = function (self, x, y, z)
		self.super.initialize(self, x, y)
		self.z = z
	end,

	point = function (self)
		return self.x, self.y, self.z
	end,
}

-- basic class tests
p1 = Point.new(10, 20)
eq(tostring(p1), "#<Class Point>")
eq(p1.x, 10)
eq(p1.y, 20)
x, y = p1:point()
eq(x, 10)
eq(y, 20)

-- metamethod (overload)
p2 = Point.new(20, 5)
pp = p1 + p2
eq(pp.x, 30)
eq(pp.y, 25)


-- inherited class tests
p3 = Point3D.new(10, 20, 30)
eq(tostring(p3), "#<Class Point3D>")
eq(p3.x, 10)
eq(p3.y, 20)
eq(p3.z, 30)
x, y, z = p3:point()
eq(x, 10)
eq(y, 20)
eq(z, 30)

pp = p1 + p3
eq(pp.x, 20)
eq(pp.y, 40)

pp = p3 + p1
eq(pp.x, 20)
eq(pp.y, 40)

-- eq
--[[
p4 = Point.new(10, 10)
p5 = Point3D.new(10, 10)
eq(p4 == p5, true)
]]
