
require "SandBox"
require "Test"

FILE_NAME = "lua_t-sandbox_test"

function createTmpFile(str)
	local tmp = os.tmpname()
	local fp = io.open(tmp, "w")
	fp:write(str)
	fp:close()
	return tmp
end

sandbox = SandBox.new()

eq(sandbox.pathComponents("AI.lua"):join(), "AI.lua")
eq(sandbox.pathComponents("AI/USER_AI/AI.lua"):join(), "AI, USER_AI, AI.lua")
eq(sandbox.pathComponents("/AI/USER_AI/AI.lua"):join(), ", AI, USER_AI, AI.lua")

---- ちゃんとよべるか。せっとできるか

local tmp = createTmpFile([[
function Test (foo)
	return foo
end
]])

local env = sandbox:load(tmp)
eq(env.Test("foo"), "foo")
env.hoge = function () return "test" end
eq(env.hoge(), "test")

---- ちゃんと関数コピーされてる?

function test ()
	return "aaa"
end

local tmp = createTmpFile([[
function Test ()
	return test()
end
]])

local env1 = sandbox:load(tmp)

eq(env1.Test(), "aaa")

---- 関数おきかえ
os.remove("/tmp/"..FILE_NAME)


local tmp = createTmpFile([[
function w ()
	f = io.open(FILE_NAME, "w")
	f:write("aaa")
	f:close()
end
]])
local env2 = sandbox:load(tmp)

env2.io.open = function (filename, mode)
	return io.open("/tmp/"..filename, mode)
end

os.remove("/tmp/"..FILE_NAME)
env2.w()
f = io.open("/tmp/"..FILE_NAME, "r")
eq(f:read("*a"), [[aaa]])
f:close()


---- ディレクトリ制限
os.remove("/tmp/"..FILE_NAME)

local tmp = createTmpFile([[
function w ()
	f = io.open(FILE_NAME, "w")
	f:write("aaa")
	f:close()
end
]])
local env3 = sandbox:load(tmp, "/tmp")
env2.w()

f = io.open("/tmp/"..FILE_NAME, "r")
eq(f:read("*a"), [[aaa]])
f:close()
